/**
 * Dynamo Droplet-like functionality for JSP.
 *
 * This package is early beta, not for commercial use! :-)
 * Read: The interfaces and classes in this package (and subpackages) will be
 * developed and modified for a while.
 *
 */
package com.twelvemonkeys.servlet.jsp.droplet;